use std::f64::consts::PI;
use mpi::collective::SystemOperation;
use mpi::traits::*;

fn main() {
    let universe = mpi::initialize().unwrap();
    let world = universe.world();

    let size = world.size();
    let rank = world.rank();

    let support_points = 100;
    let a = 0.0;
    let b = 1.0;
    let N = (b - a) / size as f64;

    let f = |x: f64| 1.0 / (1.0 + x.powi(2)) as f64; // function
    let n = (support_points / size) as u32; // fewer ranks should also use an appropriate amount of support points; the support points are distributed among the ranks
    let partial_integral = trapezoidal_rule(f, a + rank as f64 * N, a + (rank + 1) as f64 * N, n);
    println!("Rank {}: Partial integral is {} on [{},{}]", rank, partial_integral, (a + rank as f64 * N), (a + (rank + 1) as f64 * N));

    world.barrier();

    let mut integral = 0.0;

    world.all_reduce_into(&partial_integral, &mut integral, SystemOperation::sum());

    if rank == 0 {
        println!("The approximated integral is {}", integral);
        println!("The error is |{:.7} - {:.7}| = {}", (PI / 4.0), integral, (integral - (PI / 4.0)).abs());
    }
}

fn trapezoidal_rule(f: fn(f64) -> f64, a: f64, b: f64, n: u32) -> f64 {
    let h = (b - a) / n as f64;
    let mut result = 0.5 * (f(a) + f(b));

    for i in 1..n {
        let x_i = a + i as f64 * h;
        result += f(x_i);
    }

    h * result
}

#[cfg(test)]
mod tests {
    use std::f64::consts::PI;
    use crate::trapezoidal_rule;

    #[test]
    fn should_return_0_trapezoid_rule() {
        let expected = 0.0;
        let actual = trapezoidal_rule(|_x| 0.0, 0.0, 10.0, 1);

        assert_eq!(expected, actual, "Integral is {} but should be {}", actual, expected);
    }

    #[test]
    fn should_return_1_trapezoid_rule() {
        let expected = 1.0;
        let actual = trapezoidal_rule(|_x| 1.0, 0.0, 1.0, 1);

        assert_eq!(expected, actual, "Integral is {} but should be {}", actual, expected);
    }

    #[test]
    fn should_return_pi_trapezoid_rule() {
        let expected = PI / 4.0;
        let actual = trapezoidal_rule(|x| 1.0 / (1.0 + x.powi(2)), 0.0, 1.0, 10);

        let tolerance = 1e-3;

        assert!((expected - actual).abs() < tolerance, "Integral is {} but should be {}", actual, expected);
    }
}
